@@ -1,13 +1,13 @@
 /*
  * gedit-bookmarks-plugin.c - Bookmarking for gedit
- * 
+ *
  * Copyright (C) 2008 Jesse van den Kieboom
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
  * any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -45,10 +45,19 @@
 #define INSERT_DATA_KEY "GeditBookmarksInsertData"
 #define METADATA_ATTR "metadata::gedit-bookmarks"
 
+#define MESSAGE_OBJECT_PATH "/plugins/bookmarks"
+#define BUS_CONNECT(bus, name, data) gedit_message_bus_connect(bus, MESSAGE_OBJECT_PATH, #name, (GeditMessageCallback)  message_##name##_cb, data, NULL)
+
 typedef struct
 {
-	gint previous_line;
-	gboolean new_user_action;
+	GtkSourceMark *bookmark;
+	GtkTextMark *mark;
+} InsertTracker;
+
+typedef struct
+{
+	GSList *trackers;
+	guint user_action;
 } InsertData;
 
 static void update_background_color		(GeditView   *view);
@@ -67,31 +76,29 @@ static void on_insert_text_before		(GtkTextBuffer *buffer,
 						 gint		len,
 						 InsertData    *data);
 
-static void on_insert_text_after		(GtkTextBuffer *buffer,
-						 GtkTextIter   *location,
-						 gchar         *text,
-						 gint		len,
-						 InsertData    *data);
-
 static void on_begin_user_action		(GtkTextBuffer *buffer,
 						 InsertData    *data);
 
 static void on_end_user_action			(GtkTextBuffer *buffer,
 						 InsertData    *data);
 
-static void on_toggle_bookmark_activate 	(GtkAction   *action, 
+static void on_toggle_bookmark_activate 	(GtkAction   *action,
 						 GeditWindow *window);
-static void on_next_bookmark_activate 		(GtkAction   *action, 
+static void on_next_bookmark_activate 		(GtkAction   *action,
 						 GeditWindow *window);
-static void on_previous_bookmark_activate 	(GtkAction   *action, 
+static void on_previous_bookmark_activate 	(GtkAction   *action,
 						 GeditWindow *window);
-static void on_tab_added 			(GeditWindow *window, 
-						 GeditTab    *tab, 
+static void on_tab_added 			(GeditWindow *window,
+						 GeditTab    *tab,
 						 GeditPlugin *plugin);
-static void on_tab_removed 			(GeditWindow *window, 
-						 GeditTab    *tab, 
+static void on_tab_removed 			(GeditWindow *window,
+						 GeditTab    *tab,
 						 GeditPlugin *plugin);
-			  
+
+static void add_bookmark (GtkSourceBuffer *buffer, GtkTextIter *iter);
+static void remove_bookmark (GtkSourceBuffer *buffer, GtkTextIter *iter);
+static void toggle_bookmark (GtkSourceBuffer *buffer, GtkTextIter *iter);
+
 typedef struct
 {
 	GtkActionGroup * action_group;
@@ -128,7 +135,7 @@ free_insert_data (InsertData *data)
 
 static GtkActionEntry const action_entries[] = {
 	{"ToggleBookmark", NULL, N_("Toggle Bookmark"), "<Control><Alt>B",
-	 N_("Toggle bookmark status of the current line"), 
+	 N_("Toggle bookmark status of the current line"),
 	 G_CALLBACK (on_toggle_bookmark_activate)},
 	{"NextBookmark", NULL, N_("Goto Next Bookmark"), "<Control>B",
 	 N_("Goto the next bookmark"),
@@ -160,18 +167,18 @@ install_menu (GeditWindow *window)
 
 	manager = gedit_window_get_ui_manager (window);
 	data->action_group = gtk_action_group_new ("GeditBookmarksPluginActions");
-	
+
 	gtk_action_group_set_translation_domain (data->action_group,
 						 GETTEXT_PACKAGE);
-	
-	gtk_action_group_add_actions (data->action_group, 
+
+	gtk_action_group_add_actions (data->action_group,
 				      action_entries,
 				      G_N_ELEMENTS (action_entries),
 				      window);
-				      
+
 	gtk_ui_manager_insert_action_group (manager, data->action_group, -1);
 	data->ui_id = gtk_ui_manager_add_ui_from_string (manager, uidefinition, -1, &error);
-	
+
 	if (!data->ui_id)
 	{
 		g_warning ("Could not load UI: %s", error->message);
@@ -184,12 +191,12 @@ uninstall_menu (GeditWindow *window)
 {
 	WindowData *data = BOOKMARKS_DATA (window);
 	GtkUIManager *manager;
-	
+
 	manager = gedit_window_get_ui_manager (window);
-	
+
 	gtk_ui_manager_remove_ui (manager, data->ui_id);
 	gtk_ui_manager_remove_action_group (manager, data->action_group);
-	
+
 	g_object_unref (data->action_group);
 }
 
@@ -200,25 +207,24 @@ disable_bookmarks (GeditView *view)
 	GtkTextIter end;
 	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
 	gpointer data;
-	
+
 	gtk_source_view_set_show_line_marks (GTK_SOURCE_VIEW (view), FALSE);
 
 	gtk_text_buffer_get_bounds (buffer, &start, &end);
-	gtk_source_buffer_remove_source_marks (GTK_SOURCE_BUFFER (buffer), 
-					       &start, 
-					       &end, 
+	gtk_source_buffer_remove_source_marks (GTK_SOURCE_BUFFER (buffer),
+					       &start,
+					       &end,
 					       BOOKMARK_CATEGORY);
 
 	g_signal_handlers_disconnect_by_func (buffer, on_style_scheme_notify, view);
 	g_signal_handlers_disconnect_by_func (buffer, on_delete_range, NULL);
-	
+
 	data = g_object_get_data (G_OBJECT (buffer), INSERT_DATA_KEY);
 
 	g_signal_handlers_disconnect_by_func (buffer, on_insert_text_before, data);
-	g_signal_handlers_disconnect_by_func (buffer, on_insert_text_after, data);
 	g_signal_handlers_disconnect_by_func (buffer, on_begin_user_action, data);
 	g_signal_handlers_disconnect_by_func (buffer, on_end_user_action, data);
-	
+
 	g_object_set_data (G_OBJECT (buffer), INSERT_DATA_KEY, NULL);
 }
 
@@ -260,7 +266,7 @@ enable_bookmarks (GeditView   *view,
 		g_object_unref (pixbuf);
 
 		gtk_source_view_set_show_line_marks (GTK_SOURCE_VIEW (view), TRUE);
-		
+
 		g_signal_connect (buffer,
 				  "notify::style-scheme",
 				  G_CALLBACK (on_style_scheme_notify),
@@ -270,13 +276,11 @@ enable_bookmarks (GeditView   *view,
 				        "delete-range",
 				        G_CALLBACK (on_delete_range),
 				        NULL);
-				        
-		data = g_slice_new (InsertData);
-		data->new_user_action = FALSE;
-		data->previous_line = -1;
-		
-		g_object_set_data_full (G_OBJECT (buffer), 
-					INSERT_DATA_KEY, 
+
+		data = g_slice_new0 (InsertData);
+
+		g_object_set_data_full (G_OBJECT (buffer),
+					INSERT_DATA_KEY,
 					data,
 					(GDestroyNotify) free_insert_data);
 
@@ -284,16 +288,12 @@ enable_bookmarks (GeditView   *view,
 				  "insert-text",
 				  G_CALLBACK (on_insert_text_before),
 				  data);
-				  
-		g_signal_connect_after (buffer,
-				        "insert-text",
-				        G_CALLBACK (on_insert_text_after),
-				        data);
-				        
+
 		g_signal_connect (buffer,
 				  "begin-user-action",
 				  G_CALLBACK (on_begin_user_action),
 				  data);
+
 		g_signal_connect (buffer,
 				  "end-user-action",
 				  G_CALLBACK (on_end_user_action),
@@ -314,28 +314,28 @@ load_bookmarks (GeditView *view,
 	GtkTextIter iter;
 	gint tot_lines;
 	gint i;
-	
+
 	buf = GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view)));
-	
+
 	gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (buf), &iter);
 	tot_lines = gtk_text_iter_get_line (&iter);
-	
+
 	for (i = 0; bookmarks != NULL && bookmarks[i] != NULL; i++)
 	{
 		gint line;
-		
+
 		line = atoi (bookmarks[i]);
 
 		if (line >= 0 && line < tot_lines)
 		{
 			GSList *marks;
-		
+
 			gtk_text_buffer_get_iter_at_line (GTK_TEXT_BUFFER (buf),
 							  &iter, line);
-			
+
 			marks = gtk_source_buffer_get_source_marks_at_iter (buf, &iter,
 									    BOOKMARK_CATEGORY);
-			
+
 			if (marks == NULL)
 			{
 				/* Add new bookmark */
@@ -364,7 +364,7 @@ load_bookmark_metadata (GeditView *view)
 	if (bookmarks_attr != NULL)
 	{
 		gchar **bookmarks;
-	
+
 		bookmarks = g_strsplit (bookmarks_attr, ",", -1);
 		g_free (bookmarks_attr);
 
@@ -374,6 +374,276 @@ load_bookmark_metadata (GeditView *view)
 	}
 }
 
+typedef gboolean (*IterSearchFunc)(GtkSourceBuffer *buffer, GtkTextIter *iter, const gchar *category);
+typedef void (*CycleFunc)(GtkTextBuffer *buffer, GtkTextIter *iter);
+
+static void
+goto_bookmark (GeditWindow    *window,
+               GtkSourceView  *view,
+               GtkTextIter    *iter,
+	       IterSearchFunc  func,
+	       CycleFunc       cycle_func)
+{
+	GtkTextBuffer *buffer;
+	GtkTextIter at;
+	GtkTextMark *insert;
+	GtkTextIter end;
+
+	if (view == NULL)
+	{
+		view = GTK_SOURCE_VIEW (gedit_window_get_active_view (window));
+	}
+
+	if (view == NULL)
+	{
+		return;
+	}
+
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
+
+	if (iter == NULL)
+	{
+		gtk_text_buffer_get_iter_at_mark (buffer,
+		                                  &at,
+		                                  gtk_text_buffer_get_insert (buffer));
+	}
+	else
+	{
+		at = *iter;
+	}
+
+	/* Move the iter to the beginning of the line, where the bookmarks are */
+	gtk_text_iter_set_line_offset (&at, 0);
+
+	/* Try to find the next bookmark */
+	if (!func (GTK_SOURCE_BUFFER (buffer), &at, BOOKMARK_CATEGORY))
+	{
+		GSList *marks;
+
+		/* cycle through */
+		cycle_func (buffer, &at);
+		gtk_text_iter_set_line_offset (&at, 0);
+
+		marks = gtk_source_buffer_get_source_marks_at_iter (GTK_SOURCE_BUFFER (buffer),
+		                                                    &at,
+		                                                    BOOKMARK_CATEGORY);
+
+		if (!marks && !func (GTK_SOURCE_BUFFER (buffer), &at, BOOKMARK_CATEGORY))
+		{
+			return;
+		}
+
+		g_slist_free (marks);
+	}
+
+	end = at;
+
+	if (!gtk_text_iter_forward_visible_line (&end))
+	{
+		gtk_text_buffer_get_end_iter (buffer, &end);
+	}
+	else
+	{
+		gtk_text_iter_backward_char (&end);
+	}
+
+	gtk_text_buffer_select_range (buffer, &at, &end);
+	gtk_text_view_scroll_to_iter (GTK_TEXT_VIEW (view), &at, 0.3, FALSE, 0, 0);
+}
+
+static void
+message_get_view_iter (GeditWindow    *window,
+                       GeditMessage   *message,
+                       GtkSourceView **view,
+                       GtkTextIter    *iter)
+{
+	if (gedit_message_has_key (message, "view"))
+	{
+		gedit_message_get (message, "view", view, NULL);
+	}
+	else
+	{
+		*view = GTK_SOURCE_VIEW (gedit_window_get_active_view (window));
+	}
+
+	if (!*view)
+	{
+		return;
+	}
+
+	if (gedit_message_has_key (message, "iter"))
+	{
+		gedit_message_get (message, "iter", iter, NULL);
+	}
+	else
+	{
+		GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (*view));
+		gtk_text_buffer_get_iter_at_mark (buffer,
+		                                  iter,
+		                                  gtk_text_buffer_get_insert (buffer));
+	}
+}
+
+static void
+message_toggle_cb (GeditMessageBus *bus,
+                   GeditMessage    *message,
+                   GeditWindow     *window)
+{
+	GtkSourceView *view = NULL;
+	GtkTextIter iter;
+
+	message_get_view_iter (window, message, &view, &iter);
+
+	if (!view)
+	{
+		return;
+	}
+
+	toggle_bookmark (GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view))),
+	                 &iter);
+}
+
+static void
+message_add_cb (GeditMessageBus *bus,
+                GeditMessage    *message,
+                GeditWindow     *window)
+{
+	GtkSourceView *view = NULL;
+	GtkTextIter iter;
+
+	message_get_view_iter (window, message, &view, &iter);
+
+	if (!view)
+	{
+		return;
+	}
+
+	add_bookmark (GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view))),
+	              &iter);
+}
+
+static void
+message_remove_cb (GeditMessageBus *bus,
+                   GeditMessage    *message,
+                   GeditWindow     *window)
+{
+	GtkSourceView *view = NULL;
+	GtkTextIter iter;
+
+	message_get_view_iter (window, message, &view, &iter);
+
+	if (!view)
+	{
+		return;
+	}
+
+	remove_bookmark (GTK_SOURCE_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (view))),
+	                 &iter);
+}
+
+static void
+message_goto_next_cb (GeditMessageBus *bus,
+                      GeditMessage    *message,
+                      GeditWindow     *window)
+{
+	GtkSourceView *view = NULL;
+	GtkTextIter iter;
+
+	message_get_view_iter (window, message, &view, &iter);
+
+	if (!view)
+	{
+		return;
+	}
+
+	goto_bookmark (window,
+	               view,
+	               &iter,
+	               gtk_source_buffer_forward_iter_to_source_mark,
+	               gtk_text_buffer_get_start_iter);
+}
+
+static void
+message_goto_previous_cb (GeditMessageBus *bus,
+                          GeditMessage    *message,
+                          GeditWindow     *window)
+{
+	GtkSourceView *view = NULL;
+	GtkTextIter iter;
+
+	message_get_view_iter (window, message, &view, &iter);
+
+	if (!view)
+	{
+		return;
+	}
+
+	goto_bookmark (window,
+	               view,
+	               &iter,
+	               gtk_source_buffer_backward_iter_to_source_mark,
+	               gtk_text_buffer_get_end_iter);
+}
+
+static void
+install_messages (GeditWindow *window)
+{
+	GeditMessageBus *bus = gedit_window_get_message_bus (window);
+
+	gedit_message_bus_register (bus,
+	                            MESSAGE_OBJECT_PATH,
+	                           "toggle",
+	                            2,
+	                            "view", GTK_TYPE_SOURCE_VIEW,
+	                            "iter", GTK_TYPE_TEXT_ITER,
+				    NULL);
+
+	gedit_message_bus_register (bus,
+	                            MESSAGE_OBJECT_PATH,
+	                            "add",
+	                            2,
+	                            "view", GTK_TYPE_SOURCE_VIEW,
+	                            "iter", GTK_TYPE_TEXT_ITER,
+				    NULL);
+
+	gedit_message_bus_register (bus,
+	                            MESSAGE_OBJECT_PATH,
+	                            "remove",
+	                            2,
+	                            "view", GTK_TYPE_SOURCE_VIEW,
+	                            "iter", GTK_TYPE_TEXT_ITER,
+				    NULL);
+
+	gedit_message_bus_register (bus,
+	                            MESSAGE_OBJECT_PATH,
+	                            "goto_next",
+	                            2,
+	                            "view", GTK_TYPE_SOURCE_VIEW,
+	                            "iter", GTK_TYPE_TEXT_ITER,
+				    NULL);
+
+	gedit_message_bus_register (bus,
+	                            MESSAGE_OBJECT_PATH,
+	                            "goto_previous",
+	                            2,
+	                            "view", GTK_TYPE_SOURCE_VIEW,
+	                            "iter", GTK_TYPE_TEXT_ITER,
+				    NULL);
+
+	BUS_CONNECT (bus, toggle, window);
+	BUS_CONNECT (bus, add, window);
+	BUS_CONNECT (bus, remove, window);
+	BUS_CONNECT (bus, goto_next, window);
+	BUS_CONNECT (bus, goto_previous, window);
+}
+
+static void
+uninstall_messages (GeditWindow *window)
+{
+	GeditMessageBus *bus = gedit_window_get_message_bus (window);
+	gedit_message_bus_unregister_all (bus, MESSAGE_OBJECT_PATH);
+}
+
 static void
 impl_activate (GeditPlugin *plugin,
 	       GeditWindow *window)
@@ -389,7 +659,7 @@ impl_activate (GeditPlugin *plugin,
 				WINDOW_DATA_KEY,
 				data,
 				(GDestroyNotify) free_window_data);
-	
+
 	views = gedit_window_get_views (window);
 	for (item = views; item != NULL; item = item->next)
 	{
@@ -406,6 +676,7 @@ impl_activate (GeditPlugin *plugin,
 			  G_CALLBACK (on_tab_removed), plugin);
 
 	install_menu (window);
+	install_messages (window);
 }
 
 static void
@@ -418,19 +689,19 @@ save_bookmark_metadata (GeditView *view)
 	gboolean first = TRUE;
 
 	buf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
-	
+
 	gtk_text_buffer_get_start_iter (buf, &iter);
-	
+
 	string = g_string_new (NULL);
-	
+
 	while (gtk_source_buffer_forward_iter_to_source_mark (GTK_SOURCE_BUFFER (buf),
 							      &iter,
 							      BOOKMARK_CATEGORY))
 	{
 		gint line;
-		
+
 		line = gtk_text_iter_get_line (&iter);
-		
+
 		if (!first)
 		{
 			g_string_append_printf (string, ",%d", line);
@@ -449,7 +720,7 @@ save_bookmark_metadata (GeditView *view)
 	}
 	else
 		val = g_string_free (string, FALSE);
-	
+
 	gedit_document_set_metadata (GEDIT_DOCUMENT (buf), METADATA_ATTR,
 				     val, NULL);
 
@@ -467,19 +738,20 @@ impl_deactivate	(GeditPlugin *plugin,
 	gedit_debug (DEBUG_PLUGINS);
 
 	uninstall_menu (window);
-	
+	uninstall_messages (window);
+
 	views = gedit_window_get_views (window);
 
 	for (item = views; item != NULL; item = item->next)
 	{
 		disable_bookmarks (GEDIT_VIEW (item->data));
 	}
-	
+
 	g_list_free (views);
-	
+
 	data = BOOKMARKS_DATA (window);
 	g_return_if_fail (data != NULL);
-	
+
 	g_signal_handlers_disconnect_by_func (window, on_tab_added, plugin);
 	g_signal_handlers_disconnect_by_func (window, on_tab_removed, plugin);
 
@@ -505,19 +777,19 @@ update_background_color (GeditView *view)
 	GtkSourceStyle *style;
 	GtkSourceStyleScheme *scheme;
 	GtkTextBuffer *buffer;
-	
+
 	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
-	
+
 	scheme = gtk_source_buffer_get_style_scheme (GTK_SOURCE_BUFFER (buffer));
 	style = gtk_source_style_scheme_get_style (scheme, "search-match");
-	
+
 	if (style)
 	{
 		gboolean bgset;
 		gchar *bg;
-		
+
 		g_object_get (style, "background-set", &bgset, "background", &bg, NULL);
-		
+
 		if (bgset)
 		{
 			GdkColor color;
@@ -526,17 +798,17 @@ update_background_color (GeditView *view)
 								      BOOKMARK_CATEGORY,
 								      &color);
 			g_free (bg);
-			
+
 			return;
 		}
 	}
-	
+
 	gtk_source_view_set_mark_category_background (source_view,
 						      BOOKMARK_CATEGORY,
 						      NULL);
 }
 
-static void 
+static void
 on_style_scheme_notify (GObject     *object,
 			GParamSpec  *pspec,
 			GeditView   *view)
@@ -544,7 +816,7 @@ on_style_scheme_notify (GObject     *object,
 	update_background_color (view);
 }
 
-static void 
+static void
 on_delete_range (GtkTextBuffer *buffer,
 	     	 GtkTextIter   *start,
 	     	 GtkTextIter   *end,
@@ -555,18 +827,18 @@ on_delete_range (GtkTextBuffer *buffer,
 	GSList *item;
 
 	iter = *start;
-	
+
 	/* move to start of line */
 	gtk_text_iter_set_line_offset (&iter, 0);
-	
+
 	/* remove any bookmarks that are collapsed on each other due to this */
-	marks = gtk_source_buffer_get_source_marks_at_iter (GTK_SOURCE_BUFFER (buffer), 
+	marks = gtk_source_buffer_get_source_marks_at_iter (GTK_SOURCE_BUFFER (buffer),
 							    &iter,
 							    BOOKMARK_CATEGORY);
 
 	if (marks == NULL)
 		return;
-	
+
 	/* remove all but the first mark */
 	for (item = marks->next; item; item = item->next)
 		gtk_text_buffer_delete_mark (buffer, GTK_TEXT_MARK (item->data));
@@ -578,79 +850,187 @@ static void
 on_begin_user_action (GtkTextBuffer *buffer,
 		      InsertData    *data)
 {
-	data->new_user_action = TRUE;
+	++data->user_action;
 }
 
 static void
 on_end_user_action (GtkTextBuffer *buffer,
 		    InsertData    *data)
 {
-	data->previous_line = -1;
+	GSList *item;
+
+	if (--data->user_action > 0)
+	{
+		return;
+	}
+
+	/* Remove trackers */
+	for (item = data->trackers; item; item = g_slist_next (item))
+	{
+		InsertTracker *tracker = item->data;
+		GtkTextIter curloc;
+		GtkTextIter newloc;
+
+		/* Move the category to the line where the mark now is */
+		gtk_text_buffer_get_iter_at_mark (buffer,
+		                                  &curloc,
+		                                  GTK_TEXT_MARK (tracker->bookmark));
+
+		gtk_text_buffer_get_iter_at_mark (buffer,
+		                                  &newloc,
+		                                  tracker->mark);
+
+		if (gtk_text_iter_get_line (&curloc) != gtk_text_iter_get_line (&newloc))
+		{
+			gtk_text_iter_set_line_offset (&newloc, 0);
+			gtk_text_buffer_move_mark (buffer,
+			                           GTK_TEXT_MARK (tracker->bookmark),
+			                           &newloc);
+		}
+
+		gtk_text_buffer_delete_mark (buffer, tracker->mark);
+		g_slice_free (InsertTracker, tracker);
+	}
+
+	g_slist_free (data->trackers);
+	data->trackers = NULL;
 }
 
-static void 
+static void
+add_tracker (GtkTextBuffer *buffer,
+             GtkTextIter   *iter,
+             GtkSourceMark *bookmark,
+             InsertData    *data)
+{
+	GSList *item;
+	InsertTracker *tracker;
+
+	for (item = data->trackers; item; item = g_slist_next (item))
+	{
+		tracker = item->data;
+
+		if (tracker->bookmark == bookmark)
+		{
+			return;
+		}
+	}
+
+	tracker = g_slice_new (InsertTracker);
+	tracker->bookmark = bookmark;
+	tracker->mark = gtk_text_buffer_create_mark (buffer,
+	                                             NULL,
+	                                             iter,
+	                                             FALSE);
+
+	data->trackers = g_slist_prepend (data->trackers, tracker);
+}
+
+static void
 on_insert_text_before (GtkTextBuffer *buffer,
 		       GtkTextIter   *location,
 		       gchar         *text,
 		       gint	      len,
 		       InsertData    *data)
 {
-	if (data->new_user_action && !gtk_text_iter_starts_line (location))
-	{
-		data->previous_line = -1;
-	}
-	else if (data->new_user_action)
+	if (gtk_text_iter_starts_line (location))
 	{
 		GSList *marks;
 		marks = gtk_source_buffer_get_source_marks_at_iter (GTK_SOURCE_BUFFER (buffer),
-							    	    location,
-							    	    BOOKMARK_CATEGORY);
+		                                                    location,
+		                                                    BOOKMARK_CATEGORY);
 
-		if (marks == NULL)
+		if (marks != NULL)
 		{
-			data->previous_line = -1;
-		}
-		else
-		{
-			data->previous_line = gtk_text_iter_get_line (location);
+			add_tracker (buffer, location, marks->data, data);
 			g_slist_free (marks);
 		}
-		
-		data->new_user_action = FALSE;
 	}
 }
 
-static void 
-on_insert_text_after (GtkTextBuffer *buffer,
-		      GtkTextIter   *location,
-		      gchar         *text,
-		      gint	     len,
-		      InsertData    *data)
+static GtkSourceMark *
+get_bookmark_and_iter (GtkSourceBuffer *buffer,
+                       GtkTextIter     *iter,
+                       GtkTextIter     *start)
 {
-	gint current;
-	
-	if (data->previous_line == -1)
-		return;
-	
-	current = gtk_text_iter_get_line (location);
+	GSList *marks;
+	GtkSourceMark *ret = NULL;
 
-	if (current != data->previous_line)
+	if (!iter)
 	{
-		GtkTextIter iter = *location;
-		GSList *marks;
-		GSList *item;
-		
-		gtk_text_iter_set_line_offset (&iter, 0);
-		
-		marks = gtk_source_buffer_get_source_marks_at_line (GTK_SOURCE_BUFFER (buffer),
-								    data->previous_line,
-								    BOOKMARK_CATEGORY);
+		gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (buffer),
+		                                  start,
+		                                  gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (buffer)));
+	}
+	else
+	{
+		*start = *iter;
+	}
 
-		for (item = marks; item; item = item->next)
-			gtk_text_buffer_move_mark (buffer, GTK_TEXT_MARK (item->data), &iter);
+	gtk_text_iter_set_line_offset (start, 0);
 
-		g_slist_free (marks);
-		data->previous_line = current;
+	marks = gtk_source_buffer_get_source_marks_at_iter (buffer,
+							    start,
+							    BOOKMARK_CATEGORY);
+
+	if (marks != NULL)
+	{
+		ret = GTK_SOURCE_MARK (marks->data);
+	}
+
+	g_slist_free (marks);
+	return ret;
+}
+
+static void
+remove_bookmark (GtkSourceBuffer *buffer,
+                 GtkTextIter     *iter)
+{
+	GtkTextIter start;
+	GtkSourceMark *bookmark;
+
+	bookmark = get_bookmark_and_iter (buffer, iter, &start);
+
+	if (bookmark != NULL)
+	{
+		gtk_text_buffer_delete_mark (GTK_TEXT_BUFFER (buffer),
+		                             GTK_TEXT_MARK (bookmark));
+	}
+}
+
+static void
+add_bookmark (GtkSourceBuffer *buffer,
+              GtkTextIter     *iter)
+{
+	GtkTextIter start;
+	GtkSourceMark *bookmark;
+
+	bookmark = get_bookmark_and_iter (buffer, iter, &start);
+
+	if (bookmark == NULL)
+	{
+		gtk_source_buffer_create_source_mark (GTK_SOURCE_BUFFER (buffer),
+						      NULL,
+						      BOOKMARK_CATEGORY,
+						      &start);
+	}
+}
+
+static void
+toggle_bookmark (GtkSourceBuffer *buffer,
+                 GtkTextIter     *iter)
+{
+	GtkTextIter start;
+	GtkSourceMark *bookmark = NULL;
+
+	bookmark = get_bookmark_and_iter (buffer, iter, &start);
+
+	if (bookmark != NULL)
+	{
+		remove_bookmark (buffer, &start);
+	}
+	else
+	{
+		add_bookmark (buffer, &start);
 	}
 }
 
@@ -658,118 +1038,30 @@ static void
 on_toggle_bookmark_activate (GtkAction   *action,
 			     GeditWindow *window)
 {
-	GeditDocument *doc = gedit_window_get_active_document (window);
-	GtkTextMark *insert;
-	GtkSourceMark *bookmark = NULL;
-	GtkTextBuffer *buffer;
-	GtkTextIter iter;
-	GSList *marks;
-
-	if (!doc)
-		return;
-	
-	buffer = GTK_TEXT_BUFFER (doc);	
-	insert = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &iter, insert);
-	
-	/* Move the iter to the beginning of the line, where the bookmarks are */
-	gtk_text_iter_set_line_offset (&iter, 0);
-	
-	marks = gtk_source_buffer_get_source_marks_at_iter (GTK_SOURCE_BUFFER (buffer), 
-							    &iter, 
-							    BOOKMARK_CATEGORY);
-	
-	if (marks)
-		bookmark = GTK_SOURCE_MARK (marks->data);
-	
-	g_slist_free (marks);
-	
-	if (bookmark)
-	{
-		/* Remove the bookmark */
-		gtk_text_buffer_delete_mark (buffer, GTK_TEXT_MARK (bookmark));
-	}
-	else
-	{
-		/* Add new bookmark */
-		gtk_source_buffer_create_source_mark (GTK_SOURCE_BUFFER (buffer), 
-						      NULL, 
-						      BOOKMARK_CATEGORY, 
-						      &iter);
-	}
-}
-
-typedef gboolean (*IterSearchFunc)(GtkSourceBuffer *buffer, GtkTextIter *iter, const gchar *category);
-typedef void (*CycleFunc)(GtkTextBuffer *buffer, GtkTextIter *iter);
-
-static void
-goto_bookmark (GeditWindow    *window,
-	       IterSearchFunc  func,
-	       CycleFunc       cycle_func)
-{
-	GeditView *view = gedit_window_get_active_view (window);
-	GeditDocument *doc = gedit_window_get_active_document (window);
-	GtkTextBuffer *buffer;
-	GtkTextMark *insert;
-	GtkTextIter iter;
-	GtkTextIter end;
-	
-	if (doc == NULL)
-		return;
-	
-	buffer = GTK_TEXT_BUFFER (doc);	
-	insert = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &iter, insert);
-	
-	/* Move the iter to the beginning of the line, where the bookmarks are */
-	gtk_text_iter_set_line_offset (&iter, 0);
-
-	/* Try to find the next bookmark */
-	if (!func (GTK_SOURCE_BUFFER (buffer), &iter, BOOKMARK_CATEGORY))
-	{
-		GSList *marks;
-		
-		/* cycle through */
-		cycle_func (buffer, &iter);
-		gtk_text_iter_set_line_offset (&iter, 0);
-		
-		marks = gtk_source_buffer_get_source_marks_at_iter (GTK_SOURCE_BUFFER (buffer),
-							    	    &iter,
-							    	    BOOKMARK_CATEGORY);
-		
-		if (!marks && !func (GTK_SOURCE_BUFFER (buffer), &iter, BOOKMARK_CATEGORY))
-			return;
-		
-		g_slist_free (marks);
-	}
-	
-	end = iter;
-		
-	if (!gtk_text_iter_forward_visible_line (&end))
-		gtk_text_buffer_get_end_iter (buffer, &end);
-	else
-		gtk_text_iter_backward_char (&end);
-	
-	gtk_text_buffer_select_range (buffer, &iter, &end);
-	gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (view), insert, 0.3, FALSE, 0, 0);
+	GtkSourceBuffer *buffer = GTK_SOURCE_BUFFER (gedit_window_get_active_document (window));
+	toggle_bookmark (buffer, NULL);
 }
 
 static void
 on_next_bookmark_activate (GtkAction   *action,
 			   GeditWindow *window)
 {
-	goto_bookmark (window, 
-		       gtk_source_buffer_forward_iter_to_source_mark,
-		       gtk_text_buffer_get_start_iter);
+	goto_bookmark (window,
+	               NULL,
+	               NULL,
+	               gtk_source_buffer_forward_iter_to_source_mark,
+	               gtk_text_buffer_get_start_iter);
 }
 
 static void
 on_previous_bookmark_activate (GtkAction   *action,
 			       GeditWindow *window)
 {
-	goto_bookmark (window, 
-		       gtk_source_buffer_backward_iter_to_source_mark,
-		       gtk_text_buffer_get_end_iter);
+	goto_bookmark (window,
+	               NULL,
+	               NULL,
+	               gtk_source_buffer_backward_iter_to_source_mark,
+	               gtk_text_buffer_get_end_iter);
 }
 
 static void
@@ -804,7 +1096,7 @@ on_tab_added (GeditWindow *window,
 
 	doc = gedit_tab_get_document (tab);
 	view = gedit_tab_get_view (tab);
-	
+
 	g_signal_connect (doc, "loaded",
 			  G_CALLBACK (on_document_loaded),
 			  view);
@@ -825,7 +1117,7 @@ on_tab_removed (GeditWindow *window,
 
 	doc = gedit_tab_get_document (tab);
 	view = gedit_tab_get_view (tab);
-	
+
 	g_signal_handlers_disconnect_by_func (doc, on_document_loaded, view);
 	g_signal_handlers_disconnect_by_func (doc, on_document_saved, view);
 
